"""
Module for XRD simulation (experimental stage)
"""

import collections
import importlib.resources
import os

import numpy as np
from monty.serialization import loadfn
from scipy.interpolate import interp1d

from pyxtal.database.element import Element

with importlib.resources.as_file(
    importlib.resources.files("pyxtal") / "database" / "atomic_scattering_params.json"
) as path:
    ATOMIC_SCATTERING_PARAMS = loadfn(path)


class XRD:
    """
    a class to compute the powder XRD.

    Args:
        crystal: ase atoms object
        wavelength: float
        max2theta: float
        per_N: int
        ncpu: int
        preferred_orientation: boolean
        march_parameter: float
    """

    def __init__(
        self,
        crystal,
        wavelength=1.54184,
        thetas=None,
        res=0.01,
        per_N=30000,
        ncpu=1,
        filename=None,
        preferred_orientation=False,
        march_parameter=None,
    ):
        if thetas is None:
            thetas = [0, 180]
        self.res = np.radians(res)
        if filename is None:
            self.wavelength = wavelength
            self.min2theta = np.radians(thetas[0])
            self.max2theta = np.radians(thetas[1])
            self.per_N = per_N
            self.ncpu = ncpu
            self.name = crystal.get_chemical_formula()
            self.preferred_orientation = preferred_orientation
            self.march_parameter = march_parameter
            self.all_dhkl(crystal)
            self.skip_hkl = self.intensity(crystal)
            self.pxrdf()
        else:
            self.load(filename)

    def save(self, filename):
        """
        savetxt file
        """
        header = f"wavelength/thetas {self.wavelength:12.6f} {np.degrees(self.min2theta):6.2f} {np.degrees(self.max2theta):6.2f}"
        np.savetxt(filename, self.pxrd, header=header)

    def load(self, filename):
        """
        Load the pxrd from txt file
        """
        fp = open(filename)
        tmp = fp.readline()
        res = tmp.split()[2:]
        self.wavelength = float(res[0])
        self.min2theta = np.radians(float(res[1]))
        self.max2theta = np.radians(float(res[2]))

        pxrd = np.loadtxt(filename)
        self.theta2 = pxrd[:, 0]
        self.d_hkls = pxrd[:, 1]
        self.xrd_intensity = pxrd[:, -1]
        hkl_labels = []
        for i in range(len(pxrd)):
            h, k, l = int(pxrd[i, 2]), int(pxrd[i, 3]), int(pxrd[i, 4])
            hkl_labels.append([{"hkl": (h, k, l), "multiplicity": 1}])
        self.hkl_labels = hkl_labels
        self.pxrd = pxrd
        self.name = filename

    def __str__(self):
        return self.by_hkl()

    def __repr__(self):
        return str(self)

    def by_hkl(self, hkl=None):
        """
        d for any give abitray [h,k,l] index
        """
        s = ""
        if hkl is None:
            id1 = self.hkl_labels
            seqs = range(len(id1))
        else:
            seqs = None
            for id, label in enumerate(self.hkl_labels):
                hkl0 = list(label[0]["hkl"])  # label['multiplicity']
                if hkl == hkl0:
                    seqs = [id]

        if seqs is not None:
            s += "  2theta     d_hkl     hkl       Intensity  Multi\n"
            for i in seqs:
                s += f"{self.theta2[i]:8.3f}  {self.d_hkls[i]:8.3f}   "
                s += "[{:2d} {:2d} {:2d}]".format(*self.hkl_labels[i][0]["hkl"])
                s += f" {100 * self.xrd_intensity[i] / max(self.xrd_intensity):8.2f} "
                s += "{:8d}\n".format(self.hkl_labels[i][0]["multiplicity"])
        else:
            s += "This hkl is not in the given 2theta range"

        return s

    def all_dhkl(self, crystal):
        """
        3x3 representation -> 1x6 (a, b, c, alpha, beta, gamma)
        """

        # rec_matrix = crystal.get_reciprocal_cell()
        rec_matrix = crystal.cell.reciprocal()
        d_max = self.wavelength / np.sin(self.min2theta / 2) / 2
        d_min = self.wavelength / np.sin(self.max2theta / 2) / 2

        # This block is to find the shortest d_hkl
        # for all basic directions (1,0,0), (0,1,0), (1,1,0), (1,-1,0)

        hkl_index = create_index()  # 2, 2, 2)
        hkl_max = np.array([1, 1, 1])

        # to fix soon
        for index in hkl_index:
            d = np.linalg.norm(np.dot(index, rec_matrix))
            multiple = int(np.ceil(1 / d / d_min))
            index *= multiple
            for i in range(len(hkl_max)):
                if hkl_max[i] < index[i]:
                    hkl_max[i] = index[i]

        h1, k1, l1 = hkl_max
        h = np.arange(-h1, h1 + 1)
        k = np.arange(-k1, k1 + 1)
        l = np.arange(-l1, l1 + 1)

        hkl = np.array(np.meshgrid(h, k, l)).transpose()
        hkl_list = np.reshape(hkl, [len(h) * len(k) * len(l), 3])
        hkl_list = hkl_list[np.where(hkl_list.any(axis=1))[0]]
        # id = int((len(hkl_list)-1)/2)
        # hkl_list = np.delete(hkl_list, int((len(hkl_list)-1)/2), axis=0)
        d_hkl = 1 / np.linalg.norm(np.dot(hkl_list, rec_matrix), axis=1)

        shortlist = np.where((d_hkl >= d_min) & (d_hkl < d_max))[0]
        d_hkl = d_hkl[shortlist]
        hkl_list = hkl_list[shortlist]
        sintheta = self.wavelength / 2 / d_hkl

        self.theta = np.arcsin(sintheta)
        self.hkl_list = np.array(hkl_list, dtype=int)
        self.d_hkl = d_hkl

    def intensity(self, crystal, TWO_THETA_TOL=1e-5, SCALED_INTENSITY_TOL=1e-5):
        """
        This function calculates all that is necessary to find the intensities.
        This scheme is similar to pymatgen
        If the number of hkl is significanly large, will automtically switch to
        the fast mode in which we only calculate the intensity and do not care
        the exact hkl families

        Args:
            TWO_THETA_TOL: tolerance to find repeating angles
            SCALED_INTENSITY_TOL: threshold for intensities
        """
        # obtain scattering parameters, atomic numbers, and occus
        # print("total number of hkl lists", len(self.hkl_list))
        # print("total number of coords:", len(crystal.get_scaled_positions()))
        # from time import time
        # t0 = time()

        N_atom, N_hkls = len(crystal), len(self.hkl_list)
        # Make sure the code don't split into too many cycles
        if self.per_N < N_atom:
            self.per_N = N_atom

        coeffs = np.zeros([N_atom, 4, 2])
        zs = np.zeros([N_atom, 1], dtype=int)
        for i, elem in enumerate(crystal.get_chemical_symbols()):
            if elem == "D":
                elem = "H"
            coeffs[i, :, :] = ATOMIC_SCATTERING_PARAMS[elem]
            zs[i] = Element(elem).z

        # A heavy calculation, Partition it to prevent the memory issue
        s2s = (np.sin(self.theta) / self.wavelength) ** 2  # M
        hkl_per_proc = int(self.per_N / N_atom)
        N_cycle = int(np.ceil(N_hkls / hkl_per_proc))
        positions = crystal.get_scaled_positions()

        if self.ncpu == 1:
            N_cycles = range(N_cycle)
            Is = get_all_intensity(N_cycles, N_atom, self.per_N, positions, self.hkl_list, s2s, coeffs, zs)
        else:
            import multiprocessing as mp

            queue = mp.Queue()
            cycle_per_cpu = int(np.ceil(N_cycle / self.ncpu))
            if False:
                Is = np.zeros([N_hkls])
                for cycle in range(cycle_per_cpu):
                    processes = []
                    for cpu in range(self.ncpu):
                        mycycle = self.ncpu * cycle + cpu
                        Start = mycycle * hkl_per_proc
                        End = min([N_hkls, Start + hkl_per_proc])
                        # print(N_cycle, mycycle, cpu, Start, End, N_hkls)
                        if Start < End:
                            p = mp.Process(
                                target=get_all_intensity_par,
                                args=(
                                    cpu,
                                    queue,
                                    [mycycle],
                                    Start,
                                    End,
                                    hkl_per_proc,
                                    positions,
                                    self.hkl_list[Start:End],
                                    s2s[Start:End],
                                    coeffs,
                                    zs,
                                ),
                            )
                            p.start()
                            processes.append(p)

                    # print("collect results")
                    unsorted_result = [queue.get() for p in processes]
                    for p in processes:
                        p.join()
                    for t in unsorted_result:
                        Is[t[1] : t[2]] += t[3]
                    # print('Done=================', cycle)

            else:
                processes = []
                for cpu in range(self.ncpu):
                    N1 = cpu * cycle_per_cpu
                    Start = N1 * hkl_per_proc
                    if cpu + 1 == self.ncpu:
                        N2 = N_cycle
                        End = N_hkls
                    else:
                        N2 = (cpu + 1) * cycle_per_cpu
                        End = N2 * hkl_per_proc

                    cycles = range(N1, N2)

                    p = mp.Process(
                        target=get_all_intensity_par,
                        args=(
                            cpu,
                            queue,
                            cycles,
                            Start,
                            End,
                            hkl_per_proc,
                            positions,
                            self.hkl_list[Start:End],
                            s2s[Start:End],
                            coeffs,
                            zs,
                        ),
                    )
                    p.start()
                    processes.append(p)

                unsorted_result = [queue.get() for p in processes]
                for p in processes:
                    p.join()

                # collect results
                Is = np.zeros([N_hkls])
                for t in unsorted_result:
                    Is[t[1] : t[2]] += t[3]

        # Lorentz polarization factor
        lfs = (1 + np.cos(2 * self.theta) ** 2) / (np.sin(self.theta) ** 2 * np.cos(self.theta))

        # Preferred orientation factor
        if self.preferred_orientation is not False:
            G = self.march_parameter
            pos = ((G * np.cos(self.theta)) ** 2 + 1 / G * np.sin(self.theta) ** 2) ** (-3 / 2)
        else:
            pos = np.ones(N_hkls)

        # Group the peaks by theta values
        _two_thetas = np.degrees(2 * self.theta)
        self.peaks = {}

        N = int((self.max2theta - self.min2theta) / self.res)
        if len(self.hkl_list) > N:
            skip_hkl = True
            refs = np.degrees(np.linspace(self.min2theta, self.max2theta, N + 1))
            dtol = np.degrees(self.res / 2)
            for ref_theta in refs:
                ids = np.where(np.abs(_two_thetas - ref_theta) < dtol)[0]
                if len(ids) > 0:
                    intensity = np.sum(Is[ids] * lfs[ids] * pos[ids])
                    self.peaks[ref_theta] = [
                        intensity,
                        self.hkl_list[ids],
                        self.d_hkl[ids[0]],
                    ]
        else:
            skip_hkl = False
            two_thetas = []
            for id in range(len(self.hkl_list)):
                hkl, d_hkl = self.hkl_list[id], self.d_hkl[id]
                # find where the scattered angles are equal
                ind = np.where(np.abs(np.subtract(two_thetas, _two_thetas[id])) < TWO_THETA_TOL)
                if len(ind[0]) > 0:
                    # append intensity, hkl plane, and thetas to lists
                    self.peaks[two_thetas[ind[0][0]]][0] += Is[id] * lfs[id] * pos[id]
                    self.peaks[two_thetas[ind[0][0]]][1].append(tuple(hkl))
                else:
                    self.peaks[_two_thetas[id]] = [
                        Is[id] * lfs[id] * pos[id],
                        [tuple(hkl)],
                        d_hkl,
                    ]
                    two_thetas.append(_two_thetas[id])

        # obtain important intensities (defined by SCALED_INTENSITY_TOL)
        # and corresponding 2*theta, hkl plane + multiplicity, and d_hkl

        max_intensity = max([v[0] for v in self.peaks.values()])
        x = []
        y = []
        hkls = []
        d_hkls = []
        count = 0
        for k in sorted(self.peaks.keys()):
            count += 1
            v = self.peaks[k]
            if skip_hkl:
                fam = {}
                fam[tuple(v[1][0])] = len(v[1])
            else:
                fam = self.get_unique_families(v[1])
            if v[0] / max_intensity * 100 > SCALED_INTENSITY_TOL:
                # print(k, v[0]/max_intensity)
                x.append(k)
                y.append(v[0])

                hkls.append([{"hkl": hkl, "multiplicity": mult} for hkl, mult in fam.items()])
                d_hkls.append(v[2])

        self.theta2 = x
        self.xrd_intensity = y
        self.hkl_labels = hkls
        self.d_hkls = d_hkls

        return skip_hkl

    def pxrdf(self):
        """
        Group the equivalent hkl planes together by 2\theta angle
        N*6 arrays, Angle, d_hkl, h, k, l, intensity
        """
        rank = range(len(self.theta2))  # np.argsort(self.theta2)
        PL = []
        last = 0
        for i in rank:
            if self.xrd_intensity[i] > 0.01:
                angle = self.theta2[i]
                if abs(angle - last) < 1e-4:
                    PL[-1][-1] += self.xrd_intensity[i]
                else:
                    PL.append(
                        [
                            angle,
                            self.d_hkls[i],
                            self.hkl_labels[i][0]["hkl"][0],
                            self.hkl_labels[i][0]["hkl"][1],
                            self.hkl_labels[i][0]["hkl"][2],
                            self.xrd_intensity[i],
                        ]
                    )
                last = angle

        PL = np.array(PL)
        PL[:, -1] = PL[:, -1] / max(PL[:, -1])
        self.pxrd = PL

    def get_unique_families(self, hkls):
        """
        Returns unique families of Miller indices. Families must be permutations
        of each other.
        Args:
            hkls ([h, k, l]): List of Miller indices.
        Returns:
            {hkl: multiplicity}: A dict with unique hkl and multiplicity.
        """

        # TODO: Definitely speed it up.
        def is_perm(hkl1, hkl2):
            h1 = np.abs(hkl1)
            h2 = np.abs(hkl2)
            return all(i == j for i, j in zip(sorted(h1), sorted(h2)))

        unique = collections.defaultdict(list)
        for hkl1 in hkls:
            found = False
            for hkl2 in unique:
                if is_perm(hkl1, hkl2):
                    found = True
                    unique[hkl2].append(hkl1)
                    break
            if not found:
                unique[hkl1].append(hkl1)

        pretty_unique = {}
        for v in unique.values():
            pretty_unique[sorted(v)[-1]] = len(v)

        return pretty_unique

    @staticmethod
    def draw_hkl(hkl):
        """
        turn negative numbers in hkl to overbar
        """

        hkl_str = []
        for i in hkl:
            if i < 0:
                label = str(int(-i))
                label = r"$\bar{" + label + "}$"
                hkl_str.append(str(label))
            else:
                hkl_str.append(str(int(i)))

        return hkl_str

    def plot_pxrd(
        self,
        filename=None,
        profile=None,
        minimum_I=0.01,
        show_hkl=True,
        fontsize=None,
        figsize=(20, 10),
        res=0.02,
        fwhm=0.1,
        ax=None,
        xlim=None,
        width=1.0,
        legend=None,
        show=False,
    ):
        """
        plot PXRD

        Args:
            filename (None): name of the xrd plot. If None, show the plot
            profile: type of peak profile
            minimum_I (0.01): the minimum intensity to include in the plot
            show_hkl (True): whether or not show hkl labels
            fontsize (None): fontsize of text in the plot
            figsize ((20, 10)): figsize
            xlim (None): the 2theta range [x_min, x_max]
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        if fontsize is not None:
            mpl.rcParams.update({"font.size": fontsize})

        if xlim is None:
            x_min, x_max = 0, np.degrees(self.max2theta)
        else:
            x_min, x_max = xlim[0], xlim[1]

        if ax is None:
            fig, axes = plt.subplots(1, 1, figsize=figsize)  # plt.figure(figsize=figsize)
            axes.set_title("PXRD of " + self.name)
        else:
            axes = ax

        if profile is None:
            dx = x_max - x_min
            for i in self.pxrd:
                axes.bar(i[0], i[-1], color="b", width=width * dx / 180)
                if i[-1] > minimum_I and x_min <= i[0] <= x_max and show_hkl:
                    label = self.draw_hkl(i[2:5])
                    axes.text(i[0] - dx / 40, i[-1], label[0] + label[1] + label[2])
        else:
            spectra = self.get_profile(method=profile, res=res, user_kwargs={"FWHM": fwhm})
            label = "Profile: " + profile if legend is None else legend
            axes.plot(spectra[0], spectra[1], label=label)
            axes.legend()

        axes.set_xlim([x_min, x_max])
        axes.set_xlabel(r"2$\Theta$ ($\lambda$=" + str(self.wavelength) + r" $\AA$)")
        axes.set_ylabel("Intensity")

        if ax is None:
            axes.grid()
            if filename is None:
                if show:
                    fig.show()
            else:
                fig.savefig(filename)
                # fig.close()
            return fig, axes
        return None

    def plotly_pxrd(
        self,
        profile="gaussian",
        minimum_I=0.01,
        res=0.02,
        FWHM=0.1,
        height=450,
        html=None,
    ):
        import plotly.graph_objects as go

        """
        interactive plot for pxrd powered by plotly
        Args:
            xrd: xrd object
            html: html filename (str)
        """

        x, y, labels = [], [], []
        for i in range(len(self.pxrd)):
            theta2, d, h, k, l, I = self.pxrd[i]
            h, k, l = int(h), int(k), int(l)
            if minimum_I < I:
                label = f"<br>2&#952;: {theta2:6.2f}<br>d: {d:6.4f}<br>"
                label += f"I: {I:6.4f}</br>hkl: ({h:d}{k:d}{l:d})"
                x.append(theta2)
                y.append(-0.1)
                labels.append(label)

        trace1 = go.Bar(
            x=x,
            y=y,
            text=labels,
            hovertemplate="%{text}",
            width=0.5,
            name="hkl indices",
        )
        if profile is None:
            fig = go.Figure(data=[trace1])
        else:
            spectra = self.get_profile(method=profile, res=res, user_kwargs={"FWHM": FWHM})
            trace2 = go.Scatter(x=spectra[0], y=spectra[1], name="Profile: " + profile)
            fig = go.Figure(data=[trace2, trace1])

        fig.update_layout(
            height=height,
            xaxis_title=f"2&#952; ({self.wavelength:.4f} &#8491;)",
            yaxis_title="Intensity",
            title="PXRD of " + self.name,
        )

        if os.environ["_"].find("jupyter") == -1:
            if html is None:
                return fig.to_html()
            else:
                fig.write_html(html)
                return None
        else:
            print("This is running on Jupyter Notebook")
            return fig

    def get_profile(self, method="gaussian", res=0.01, user_kwargs=None):
        """
        return the profile detail
        """

        return Profile(method, res, user_kwargs).get_profile(
            self.theta2,
            self.xrd_intensity,
            np.degrees(self.min2theta),
            np.degrees(self.max2theta),
        )


# ----------------------------- Profile functions ------------------------------


class Profile:
    """
    This class applies a profiling function to simulated or
    experimentally obtained XRD spectra.

    Args:
        method (str): Type of function used to profile
        res (float): resolution of the profiling array in degree
        user_kwargs (dict): The parameters for the profiling method.
    """

    def __init__(self, method="mod_pseudo-voigt", res=0.02, user_kwargs=None):
        self.method = method
        self.user_kwargs = user_kwargs
        self.res = res
        kwargs = {}

        if method == "mod_pseudo-voigt":
            _kwargs = {
                "U": 5.776410e-03,
                "V": -1.673830e-03,
                "W": 5.668770e-03,
                "A": 1.03944,
                "eta_h": 0.504656,
                "eta_l": 0.611844,
            }
        elif method in ["gaussian", "lorentzian", "pseudo-voigt"]:
            _kwargs = {"FWHM": 0.1}

        else:
            msg = method + " isn't supported."
            raise NotImplementedError(msg)

        kwargs.update(_kwargs)

        if user_kwargs is not None:
            kwargs.update(user_kwargs)

        self.kwargs = kwargs

    def get_profile(self, two_thetas, intensities, min2theta, max2theta):
        """
        Performs profiling with selected function, resolution, and parameters

        Args:
            - two_thetas: 1d float array simulated/measured 2 theta values
            - intensities: simulated/measures peaks
        """
        N = int((max2theta - min2theta) / self.res)
        px = np.linspace(min2theta, max2theta, N)
        py = np.zeros(N)
        for two_theta, intensity in zip(two_thetas, intensities):
            # print(two_theta, intensity)
            if self.method == "gaussian":
                fwhm = self.kwargs["FWHM"]
                dtheta2 = ((px - two_theta) / fwhm) ** 2
                tmp = np.exp(-4 * np.log(2) * dtheta2)
                # tmp = gaussian(two_theta, px, fwhm)

            elif self.method == "lorentzian":
                fwhm = self.kwargs["FWHM"]
                tmp = lorentzian(two_theta, px, fwhm)

            elif self.method == "pseudo-voigt":
                try:
                    fwhm_g = self.kwargs["FWHM-G"]
                    fwhm_l = self.kwargs["FWHM-L"]
                except:
                    fwhm_g = self.kwargs["FWHM"]
                    fwhm_l = self.kwargs["FWHM"]

                fwhm = (
                    fwhm_g**5
                    + 2.69269 * fwhm_g**4 * fwhm_l
                    + 2.42843 * fwhm_g**3 * fwhm_l**2
                    + 4.47163 * fwhm_g**2 * fwhm_l**3
                    + 0.07842 * fwhm_g * fwhm_l**4
                    + fwhm_l**5
                ) ** (1 / 5)
                eta = 1.36603 * fwhm_l / fwhm - 0.47719 * (fwhm_l / fwhm) ** 2 + 0.11116 * (fwhm_l / fwhm) ** 3
                tmp = pseudo_voigt(two_theta, px, fwhm, eta)

            elif self.method == "mod_pseudo-voigt":
                U = self.kwargs["U"]
                V = self.kwargs["V"]
                W = self.kwargs["W"]
                A = self.kwargs["A"]
                eta_h = self.kwargs["eta_h"]
                eta_l = self.kwargs["eta_l"]

                fwhm = np.sqrt(
                    U * np.tan(np.pi * two_theta / 2 / 180) ** 2 + V * np.tan(np.pi * two_theta / 2 / 180) + W
                )
                x = px - two_theta
                tmp = mod_pseudo_voigt(x, fwhm, A, eta_h, eta_l, N)

            py += intensity * tmp
            # print(intensity * tmp)

        py /= np.max(py)

        self.spectra = np.vstack((px, py))
        return self.spectra


# ------------------------------ Similarity between two XRDs ---------------------------------
class Similarity:
    def __init__(self, f, g, N=None, x_range=None, l=2.0, weight="cosine"):
        """
        Class to compute the similarity between two diffraction patterns
        Args:

            f: spectra1 (2D array)
            g: spectra2 (2D array)
            N: number of sampling points for the processed spectra
            x_range: the range of x values used to compute similarity ([x_min, x_max])
            l: cutoff value for shift (real)
            weight: weight function 'triangle' or 'cosine' (str)
        """

        fx, fy = f[0], f[1]
        gx, gy = g[0], g[1]
        self.l = abs(l)
        res1 = (fx[-1] - fx[0]) / len(fx)
        res2 = (gx[-1] - gx[0]) / len(gx)
        self.resolution = min([res1, res2]) / 3  # improve the resolution

        if N is None:
            self.N = int(2 * self.l / self.resolution)
        else:
            self.N = N
        self.r = np.linspace(-self.l, self.l, self.N)

        if x_range is None:
            x_min = max(np.min(fx), np.min(gx))
            x_max = min(np.max(fx), np.max(gx))
        else:
            x_min, x_max = x_range[0], x_range[1]

        self.x_range = [x_min, x_max]

        f_inter = interp1d(fx, fy, "cubic", fill_value="extrapolate")
        g_inter = interp1d(gx, gy, "cubic", fill_value="extrapolate")

        fgx_new = np.linspace(x_min, x_max, int((x_max - x_min) / self.resolution) + 1)
        fy_new = f_inter(fgx_new)
        gy_new = g_inter(fgx_new)

        self.fx, self.gx, self.fy, self.gy = fgx_new, fgx_new, fy_new, gy_new
        self.weight = weight
        if self.weight == "triangle":
            w = self.triangleFunction()
        elif self.weight == "cosine":
            w = self.cosineFunction()
        else:
            msg = self.weight + "is not supported"
            raise NotImplementedError(msg)

        Npts = len(self.fx)
        d = self.fx[1] - self.fx[0]

        self.value = similarity_calculate(self.r, w, d, Npts, self.fy, self.gy)

    def __str__(self):
        return f"The similarity between two PXRDs is {self.value:.4f}"

    def __repr__(self):
        return str(self)

    def triangleFunction(self):
        """
        Triangle function to weight correlations
        """
        w = 1 - np.abs(self.r / self.l)
        ids = np.abs(self.r) > self.l
        w[ids] = 0

        return w

    def cosineFunction(self):
        """
        cosine function to weight correlations
        """

        w = 0.5 * (np.cos(np.pi * self.r / self.l) + 1.0)
        ids = np.abs(self.r) > self.l
        w[ids] = 0

        return w

    def show(self, filename=None, fontsize=None, labels=None):
        """
        show the comparison plot

        Args:
            filename (None): name of the xrd plot. If None, show the plot
            labels [A, B]: labels of each plot
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        if labels is None:
            labels = ["profile 1", "profile 2"]
        if fontsize is not None:
            mpl.rcParams.update({"font.size": fontsize})

        fig1 = plt.figure(1, figsize=(15, 6))
        fig1.add_axes((0.1, 0.3, 0.8, 0.6))

        plt.plot(self.fx, self.fy, label=labels[0])
        plt.plot(self.fx, -self.gy, label=labels[1])
        plt.legend()

        # Residual plot
        residuals = self.gy - self.fy
        fig1.add_axes((0.1, 0.1, 0.8, 0.2))
        plt.plot(self.gx, residuals, ".r", markersize=0.5)
        plt.title(f"{self.value:6f}")

        if filename is None:
            plt.show()
        else:
            plt.savefig(filename)
            plt.close()


def mod_pseudo_voigt(x, fwhm, A, eta_h, eta_l, N):
    """
    A modified split-type pseudo-Voigt function for profiling peaks
    - Izumi, F., & Ikeda, T. (2000).
    """

    tmp = np.zeros(N)
    for xi, dx in enumerate(x):
        if dx < 0:
            A = A
            eta_l = eta_l
            eta_h = eta_h
        else:
            A = 1 / A
            eta_l = eta_h
            eta_h = eta_l

        tmp[xi] = (
            ((1 + A) * (eta_h + np.sqrt(np.pi * np.log(2)) * (1 - eta_h)))
            / (
                eta_l
                + np.sqrt(np.pi * np.log(2)) * (1 - eta_l)
                + A * (eta_h + np.sqrt(np.pi * np.log(2)) * (1 - eta_h))
            )
            * (
                eta_l * 2 / (np.pi * fwhm) * (1 + ((1 + A) / A) ** 2 * (dx / fwhm) ** 2) ** (-1)
                + (1 - eta_l)
                * np.sqrt(np.log(2) / np.pi)
                * 2
                / fwhm
                * np.exp(-np.log(2) * ((1 + A) / A) ** 2 * (dx / fwhm) ** 2)
            )
        )
    return tmp


def gaussian(theta2, alpha, fwhm):
    """
    Gaussian function for profiling peaks
    """

    tmp = ((alpha - theta2) / fwhm) ** 2
    return np.exp(-4 * np.log(2) * tmp)


def lorentzian(theta2, alpha, fwhm):
    """
    Lorentzian function for profiling peaks
    """

    tmp = 1 + 4 * ((alpha - theta2) / fwhm) ** 2
    return 1 / tmp


def pseudo_voigt(theta2, alpha, fwhm, eta):
    """
    Original Pseudo-Voigt function for profiling peaks
    - Thompson, D. E. Cox & J. B. Hastings (1986).
    """

    L = lorentzian(theta2, alpha, fwhm)
    G = gaussian(theta2, alpha, fwhm)
    return eta * L + (1 - eta) * G


def similarity_calculate(r, w, d, Npts, fy, gy):
    """
    Compute the similarity between the pair of spectra f, g
    """

    xCorrfg_w, aCorrff_w, aCorrgg_w = 0, 0, 0
    for r0, w0 in zip(r, w):
        Corrfg, Corrff, Corrgg = 0, 0, 0
        shift = int(r0 / d)
        for i in range(Npts):
            if 0 <= i + shift <= Npts - 1:
                Corrfg += fy[i] * gy[i + shift] * d
                Corrff += fy[i] * fy[i + shift] * d
                Corrgg += gy[i] * gy[i + shift] * d

        xCorrfg_w += w0 * Corrfg * d
        aCorrff_w += w0 * Corrff * d
        aCorrgg_w += w0 * Corrgg * d

    return np.abs(xCorrfg_w / np.sqrt(aCorrff_w * aCorrgg_w))


def create_index(imax=1, jmax=1, kmax=1):
    """
    shortcut to get the index
    """
    hkl_index = []
    for i in range(-imax, imax + 1):
        for j in range(-jmax, jmax + 1):
            for k in range(-kmax, kmax + 1):
                hkl = np.array([i, j, k])
                if sum(hkl * hkl) > 0:
                    hkl_index.append(hkl)
    return np.array(hkl_index).reshape([len(hkl_index), 3])


def get_intensity(positions, hkl, s2, coeffs, z):
    const = 2j * np.pi
    g_dot_rs = np.dot(positions, hkl)  # N*M
    exps = np.exp(const * g_dot_rs)  # N*M

    tmp1 = np.exp(np.einsum("ij,k->ijk", -coeffs[:, :, 1], s2))  # N*4, M
    tmp2 = np.einsum("ij,ijk->ik", coeffs[:, :, 0], tmp1)  # N*4, N*M
    sfs = np.add(-41.78214 * np.einsum("ij,j->ij", tmp2, s2), z)  # N*M, M -> N*M
    fs = np.sum(sfs * exps, axis=0)  # M

    # Final intensity values, M
    return (fs * fs.conjugate()).real


def get_all_intensity(N_cycles, N_atom, per_N, positions, hkls, s2s, coeffs, zs):
    Is = np.zeros(len(hkls))
    for i, cycle in enumerate(N_cycles):
        N1 = int(per_N * (cycle) / N_atom)
        if i + 1 == len(N_cycles):
            N2 = min([len(hkls), int(per_N * (cycle + 1) / N_atom)])
        else:
            N2 = int(per_N * (cycle + 1) / N_atom)
        hkl, s2 = hkls[N1:N2].T, s2s[N1:N2]
        Is[N1:N2] = get_intensity(positions, hkl, s2, coeffs, zs)
    return Is


def get_all_intensity_par(cpu, queue, cycles, Start, End, hkl_per_proc, positions, hkls, s2s, coeffs, zs):
    # print("proc", cpu, cycles)
    Is = np.zeros(End - Start)
    # print(cpu, Start, End)

    for i, cycle in enumerate(cycles):
        N1 = cycle * hkl_per_proc - Start
        N2 = End - Start if i + 1 == len(cycles) else N1 + hkl_per_proc
        hkl, s2 = hkls[N1:N2].T, s2s[N1:N2]
        Is[N1:N2] = get_intensity(positions, hkl, s2, coeffs, zs)
        # print('run', cpu, N1+Start, N2+Start, N1, N2)
    queue.put((cpu, Start, End, Is))


def pxrd_refine(xtal, ref_pxrd, thetas, steps=20):
    """
    Improve the lattice w.r.t the reference PXRD

    Args:
        xtal:
        ref_pxrd:
        thetas:
        steps (int):
    """
    from scipy.optimize import minimize

    def fun(x0, rep, ref_pxrd, thetas):
        rep.x[0][1:] = x0
        s = rep.to_pyxtal()
        xrd = s.get_XRD(thetas=thetas)
        pxrd = xrd.get_profile(res=0.15, user_kwargs={"FWHM": 0.25})
        sim = Similarity(ref_pxrd, pxrd, x_range=thetas).value
        return -sim

    rep = xtal.get_1D_representation()
    x0 = rep.x[0][1:]
    fun(x0, rep, ref_pxrd, thetas)
    res = minimize(
        fun,
        x0,
        args=(rep, ref_pxrd, thetas),
        method="Nelder-Mead",
        options={"maxiter": steps},
    )
    rep.x[0][1:] = res.x
    xtal = rep.to_pyxtal()
    return xtal, res.x, -res.fun


if __name__ == "__main__":
    from optparse import OptionParser

    from matplotlib import pyplot as plt
    from pymatgen.core import Structure

    from pyxtal import pyxtal
    from pyxtal.util import parse_cif

    parser = OptionParser()
    parser.add_option("-f", "--cif", dest="cif", help="cif file name")
    parser.add_option("-s", "--step", dest="step", type=int, default=20, help="steps, optional")

    (options, args) = parser.parse_args()
    thetas = [5, 50]
    data = np.loadtxt("ref.txt")
    data[:, 1] /= np.max(data[:, 1])  # Normalize the intensity
    ref_pxrd = (data[:, 0], data[:, 1])

    with open(options.cif) as f:
        lines = f.readlines()
        for l in lines:
            if l.find("smile") > 0:
                break
        smiles = []
        tmp = l.split(":")[-1].split(".")
        tmp[-1] = tmp[-1][:-1]
        for t in tmp:
            t = t.replace(" ", "")
            smiles.append(t + ".smi")

    cifs = parse_cif(options.cif)
    fig, axs = plt.subplots(
        nrows=len(cifs),
        ncols=2,
        linewidth=1,
        figsize=(12, 3 * len(cifs)),
    )
    s = pyxtal(molecular=True)
    for i, cif in enumerate(cifs):
        pmg = Structure.from_str(cif, fmt="cif")
        ax1, ax2 = axs[i, 0], axs[i, 1]
        ax1.plot(ref_pxrd[0], ref_pxrd[1], label="ref")
        ax2.plot(ref_pxrd[0], ref_pxrd[1], label="ref")
        s.from_seed(pmg, molecules=smiles)
        pxrd = s.get_XRD(thetas=thetas).get_profile(res=0.15, user_kwargs={"FWHM": 0.25})
        (s, val0, val1) = pxrd_refine(s, ref_pxrd, thetas, steps=0)
        ax1.plot(pxrd[0], pxrd[1], label=f"Init: {val1:6.3f}")

        (s, val0, val1) = pxrd_refine(s, ref_pxrd, thetas, steps=options.step)
        ax2.plot(pxrd[0], pxrd[1], label=f"Opt: {val1:6.3f}")
        pxrd = s.get_XRD(thetas=thetas).get_profile(res=0.15, user_kwargs={"FWHM": 0.25})
        ax2.legend()
        ax1.legend()

        print(i, s.lattice, val1)
    fig.savefig("pxrd_match.png", dpi=150)



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

from pyxtal.lattice import Lattice
from pyxtal.symmetry import Group, Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestSymmetry(unittest.TestCase):
    def test_from_symops_wo_grou(self):
        data = [
            (["x, y, z", "-x, y+1/2, -z"], 4, 6),
            (["x, y, z", "-x+1/2, -y, z+1/2", "-x, y, z", "x+1/2, -y, z+1/2"], 31, 155),
            (
                ["x, y, z", "-x, -y, -z", "-x+1/2, y+1/2, -z", "x+1/2, -y+1/2, z"],
                14,
                83,
            ),
            (
                [
                    "x, y, z",
                    "-x, -y, -z",
                    "-x+1/2, y+1/2, -z+1/2",
                    "x+1/2, -y+1/2, z+1/2",
                ],
                14,
                82,
            ),
        ]
        for d in data:
            (strs, spg, hall) = d
            wp = Wyckoff_position.from_symops_wo_group(strs)
            assert wp.number == spg
            assert wp.hall_number == hall

    def test_from_symops(self):
        data = [
            (["x, y, z", "-x, y+1/2, -z"], 4, 6),
            (["x, y, z", "-x+1/2, -y, z+1/2", "-x, y, z", "x+1/2, -y, z+1/2"], 31, 155),
        ]
        for d in data:
            (strs, spg, hall) = d
            G = Group(spg)
            wp = Wyckoff_position.from_symops(strs, G)
            assert wp.number == spg
            assert wp.hall_number == hall

if __name__ == "__main__":
    unittest.main()



import signal
from time import time
import numpy as np
from ase.constraints import FixSymmetry
from ase.filters import UnitCellFilter
from ase.optimize.fire import FIRE
import torchani
import logging

import os
from mace.calculators import mace_mp
_cached_mace_mp = None


def get_calculator(calculator):
    global _cached_mace_mp

    if type(calculator) is str:
        if calculator == 'ANI':
            calc = torchani.models.ANI2x().ase()
        else:
            if _cached_mace_mp is None:
                _cached_mace_mp = mace_mp(model='small', dispersion=True, device='cpu')
            calc = _cached_mace_mp
    else:
        calc = calculator

    return calc

def ASE_relax(struc, calculator, opt_cell=False, step=500, fmax=0.1, logfile=None, max_time=10.0, label='ase'):
#def ASE_relax(struc, calculator, opt_cell=False, step=500, fmax=0.1, logfile='ase.log', max_time=10.0, label='ase'):
    """
    ASE optimizer
    Args:
        struc: ase atoms object
        calculator (str): 'ANI', 'MACE'
        step: optimization steps (int)
        max_time: float (minutes)
    """

    def handler(signum, frame):
        raise TimeoutError("Optimization timed out")

    step_init = min([30, int(step/2)])
    logger = logging.getLogger()
    max_time *= 60
    timeout = int(max_time)
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout)
    #logger.info(f"{label} start calculation")
    _fmax = 1e+5

    try:
    #if True:
        calc = get_calculator(calculator)
        struc.set_calculator(calc)
        struc.set_constraint(FixSymmetry(struc))
        if opt_cell:
            ecf = UnitCellFilter(struc)
            dyn = FIRE(ecf, a=0.1, logfile=logfile) if logfile is not None else FIRE(ecf, a=0.1)
        else:
            dyn = FIRE(struc, a=0.1, logfile=logfile) if logfile is not None else FIRE(struc, a=0.1)

        # Run relaxation
        dyn.run(fmax=fmax, steps=step_init)
        forces = dyn.optimizable.get_forces()
        _fmax = np.sqrt((forces ** 2).sum(axis=1).max())

        if _fmax < 1e+3 and step > step_init:
            dyn.run(fmax=fmax, steps=step-step_init)
            forces = dyn.optimizable.get_forces()
            _fmax = np.sqrt((forces ** 2).sum(axis=1).max())
            eng = struc.get_potential_energy() / len(struc)
            if _fmax > 100:
                logger.info(f"Warning {label} big stress {eng:.2f} / {_fmax:.2f}, skip")
                struc = None
            else:
                logger.info(f"{label} Success  {eng:.2f} / {_fmax:.2f}")
        else:
            logger.info(f"Warning {label} big stress {_fmax:.2f} for 20 steps, skip")
            struc = None
        signal.alarm(0)  # Cancel the alarm if finished within time

    except TimeoutError:
        logger.info(f"Warning {label} timed out after {timeout} seconds.")
        struc = None

    except TypeError:
        logger.info(f"Warning {label} spglib error in getting the lattice")
        struc = None
        signal.alarm(0)  # Cancel the alarm if finished within time

    tag = 'False' if struc is None else 'True'
    logger.info(f"Finishing {label} {tag}")
    #signal.alarm(0)  # Cancel the alarm 
    return struc #, eng, _fmax

class ASE_optimizer:
    """
    This is a ASE optimizer to perform oragnic crystal structure optimization.
    We assume that the geometry has been well optimized by classical FF

    Args:
        struc: pyxtal object
        calculator (str): 'ANI', 'MACE'
        opt_lat (bool): to opt lattice or not
        log_file (str): output file
    """

    def __init__(self, struc, calculator='MACE', opt_lat=True, logfile=None):
        self.structure = struc
        self.calculator = get_calculator(calculator)
        self.opt_lat = opt_lat
        self.stress = None
        self.forces = None
        self.optimized = True
        self.positions = None
        self.cell = None
        self.cputime = 0
        self.logfile = logfile

    def run(self, steps=10):
        t0 = time()
        s = self.structure.to_ase(resort=False)
        s.set_constraint(FixSymmetry(s))
        s.set_calculator(self.calculator)#; print("Setup Fire")

        if not self.opt_lat:
            dyn = FIRE(s, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
        else:
            #ecf = FrechetCellFilter(s)
            ecf = UnitCellFilter(s)
            dyn = FIRE(ecf, a=0.1, logfile=self.logfile)#, force_consistent=False)
            dyn.run(fmax=0.1, steps=steps)
            self.structure.lattice.set_matrix(s.get_cell())

        positions = s.get_scaled_positions()
        #try:
        if True:
            # s.write('../1.cif', format='cif')
            count = 0
            for _i, site in enumerate(self.structure.mol_sites):
                coords0, _ = site._get_coords_and_species(first=True)
                coords1 = positions[count : count + len(site.molecule.mol)]
                for j, coor in enumerate(coords1):
                    diff = coor - coords0[j]
                    diff -= np.round(diff)
                    abs_diff = np.dot(diff, s.get_cell())
                    # print(j, coor, coords0[j], diff, np.linalg.norm(abs_diff))
                    if abs(np.linalg.norm(abs_diff)) < 2.0:
                        coords1[j] = coords0[j] + diff
                    else:
                        print(coords1[j], coords1[j], np.linalg.norm(abs_diff))
                        import sys; sys.exit()

                site.update(coords1, self.structure.lattice)
                count += len(site.molecule.mol) * site.wp.multiplicity
            self.structure.optimize_lattice()
            self.structure.energy = s.get_potential_energy()
            self.cell = s.get_cell()
            # print(self.structure.lattice)
        #except:
        #    self.structure.energy = 10000
        #    self.optimized = False
        #    print("Structure is wrong after optimization")
        s.set_calculator()
        s.set_constraint()
        self.cputime = time() - t0


if __name__ == "__main__":
    import os, warnings
    from pyxtal.db import database
    warnings.filterwarnings("ignore")

    work_dir = "tmp"
    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    db = database("pyxtal/database/test.db")
    struc = db.get_pyxtal("ACSALA")

    calc = ASE_optimizer(struc)
    print(calc.structure.lattice)
    calc.run()
    print(calc.structure.energy)
    print(calc.structure.lattice)



"""
Database class
"""

import os, time
from concurrent.futures import ProcessPoolExecutor
import logging

import numpy as np
import pymatgen.analysis.structure_matcher as sm
from ase.calculators.calculator import CalculationFailed
from ase.db import connect

from pyxtal import pyxtal
from pyxtal.util import ase2pymatgen


def setup_worker_logger(log_file):
    """
    Set up the logger for each worker process.
    """
    logging.getLogger().handlers.clear()
    logging.basicConfig(format="%(asctime)s| %(message)s",
                        filename=log_file,
                        level=logging.INFO)

def call_opt_single(p):
    """
    Optimize a single structure and log the result.

    Args:
        p (tuple): A tuple where the first element is an identifier (id), and the
                   remaining elements are the arguments to pass to `opt_single`.

    Returns:
        tuple: A tuple (id, xtal, eng) where:
               - id (int): The identifier of the structure.
               - xtal: The optimized structure.
               - eng (float): The energy of the opt_structure, or None if it failed.

    Behavior:
        This function calls `opt_single` to perform the optimization of the structure
        associated with the given id.
    """
    #logger = logging.getLogger()
    #logger.info(f"ID: {p[0]} *{sum(p[1].numIons)}")
    myid = p[0]
    xtal, eng, status = opt_single(*p)
    return myid, xtal, eng


def opt_single(id, xtal, calc, *args):
    """
    Optimize a structure using the specified calculator.

    Args:
        id (int): Identifier of the structure to be optimized.
        xtal: Crystal structure object to be optimized.
        calc (str): The calculator to use ('GULP', 'DFTB', 'VASP', 'MACE').
        *args: Additional arguments to pass to the calculator function.

    Returns:
        tuple: The result of the optimization, which typically includes:
               - xtal: The optimized structure.
               - energy (float): The energy of the optimized structure.
               - status (bool): Whether the optimization was successful.

    Raises:
        ValueError: If an unsupported calculator is specified.
    """

    if calc == 'GULP':
        return gulp_opt_single(id, xtal, *args)
    elif calc == 'DFTB':
        return dftb_opt_single(id, xtal, *args)
    elif calc == 'VASP':
        return vasp_opt_single(id, xtal, *args)
    elif calc == 'MACE':
        return mace_opt_single(id, xtal, *args)
    else:
        raise ValueError("Cannot support this calcultor", calc)


def dftb_opt_single(id, xtal, skf_dir, steps, symmetrize, criteria, kresol=0.05):
    """
    Single DFTB optimization for a given atomic xtal

    Args:
        id (int): id of the give xtal
        xtal: pyxtal instance
        skf_dir (str): path of skf files
        steps (int): number of relaxation steps
        criteria (dicts): to check if the structure
    """
    from pyxtal.interface.dftb import DFTB, DFTB_relax

    cwd = os.getcwd()
    atoms = xtal.to_ase(resort=False)
    eng = None
    stress = None
    try:
        if symmetrize:
            s = DFTB_relax(
                atoms,
                skf_dir,
                True,
                int(steps / 2),
                kresol=kresol,
                folder=".",
                scc_error=1e-5,
                scc_iter=100,
                logfile="ase.log",
            )
            stress = np.sum(s.get_stress()[:3]) / 0.006241509125883258 / 3
        else:
            my = DFTB(
                atoms,
                skf_dir,
                kresol=kresol * 1.5,
                folder=".",
                scc_error=0.1,
                scc_iter=100,
            )
            s, eng = my.run(mode="vc-relax", step=int(steps / 2))
            my = DFTB(s, skf_dir, kresol=kresol, folder=".",
                      scc_error=1e-4, scc_iter=100)
            s, eng = my.run(mode="vc-relax", step=int(steps / 2))
            s = my.struc
    except CalculationFailed:
        # This is due to covergence error in geometry optimization
        # Here we simply read the last energy
        my.calc.read_results()
        eng = my.calc.results["energy"]
        s = my.struc
    except:
        s = None
        print("Problem in DFTB Geometry optimization", id)
        xtal.to_file("bug.cif")  # ; import sys; sys.exit()
    os.chdir(cwd)

    status = False
    if s is not None:
        c = pyxtal()
        c.from_seed(s)
        if eng is None:
            eng = s.get_potential_energy() / len(s)
        else:
            eng /= len(s)

        status = process_xtal(id, xtal, eng, criteria)
        print(xtal.get_xtal_string())

        return xtal, eng, status
    else:
        return None, None, False


def vasp_opt_single(id, xtal, path, cmd, criteria):
    """
    Single VASP optimization for a given atomic xtal

    Args:
        id (int): id of the give xtal
        xtal: pyxtal instance
        path: calculation folder
        cmd: vasp command
        criteria (dicts): to check if the structure
    """
    from pyxtal.interface.vasp import optimize as vasp_opt
    cwd = os.getcwd()
    path += '/g' + str(id)
    status = False

    xtal, eng, _, error = vasp_opt(xtal,
                                   path,
                                   cmd=cmd,
                                   walltime="59m")
    if not error:
        status = process_xtal(id, xtal, eng, criteria)
    else:
        os.chdir(cwd)
    return xtal, eng, status


def gulp_opt_single(id, xtal, ff_lib, path, criteria):
    """
    Perform a single GULP optimization for a given crystal structure.

    Args:
        id (int): Identifier for the current structure.
        xtal: PyXtal instance representing the crystal to be optimized.
        ff_lib (str): Force field library for GULP, e.g., 'reaxff', 'tersoff'.
        path (str): Path to the folder where the calculation is stored.
        criteria (dict): Dictionary to check the validity of the opt_structure.

    Returns:
        tuple:
            - xtal: Optimized PyXtal instance.
            - eng (float): Energy of the optimized structure.
            - status (bool): Whether the optimization process is successful.

    Behavior:
        This function performs a GULP optimization using the force field.
        After the optimization, it checks the validity of the structure and
        attempts to remove the calculation folder if it is empty.
    """
    from pyxtal.interface.gulp import single_optimize as gulp_opt

    # Create the path for this specific structure
    path += '/g' + str(id)

    # Perform the optimization with GULP
    xtal, eng, _, error = gulp_opt(
        xtal,
        ff=ff_lib,
        label=str(id),
        path=path,
        symmetry=True,
    )

    # Default status to False, will be updated if successful
    status = False
    if not error:
        status = process_xtal(id, xtal, eng, criteria)
        try:
            os.rmdir(path)
        except:
            print("Folder is not empty", path)
    return xtal, eng, status


def mace_opt_single(id, xtal, step, criteria):
    """
    Perform a single MACE optimization for a given atomic crystal structure.

    Args:
        id (int): Identifier for the current structure.
        xtal: PyXtal instance representing the crystal structure.
        step (int): Maximum number of relaxation steps. Default is 250.
        criteria (dict): Dictionary to check the validity of the optimized structure.

    Returns:
        tuple:
            - xtal: Optimized PyXtal instance (or None if optimization failed).
            - eng (float): Energy/atom of the opt_structure (or None if it failed).
            - status (bool): Whether the optimization was successful.
    """
    from pyxtal.interface.ase_opt import ASE_relax as mace_opt

    logger = logging.getLogger()
    atoms = xtal.to_ase(resort=False)
    s = mace_opt(atoms,
                 'MACE',
                 opt_cell=True,
                 step=step,
                 max_time=9.0 * max([1, (len(atoms)/200)]),
                 label=str(id))
    if s is None:
        logger.info(f"mace_opt_single Failure {id}")
        return None, None, False

    try:
        xtal = pyxtal()
        xtal.from_seed(s)
        eng = s.get_potential_energy() / len(s)
        status = process_xtal(id, xtal, eng, criteria)
        logger.info(f"mace_opt_single Success {id}")
        return xtal, eng, status
    except:
        logger.info(f"mace_opt_single Bug {id}")
        return None, None, False


def process_xtal(id, xtal, eng, criteria):
    status = xtal.check_validity(
        criteria) if criteria is not None else True
    if status:
        header = f"{id:4d}"
        dicts = {"validity": status, "energy": eng}
        print(xtal.get_xtal_string(header=header, dicts=dicts))
    return status


def make_entry_from_pyxtal(xtal):
    """
    Generate an entry dictionary from a PyXtal object, assuming
    the SMILES and CCDC number information is provided.

    Args:
        xtal: PyXtal object (must contain the SMILES (`xtal.tag["smiles"]`)
        and CCDC number (`xtal.tag["ccdc_number"]`) in the `xtal.tag`.

    Returns:
        tuple: (ase_atoms, entry_dict, None)
            - ase_atoms: ASE Atoms object converted from the PyXtal structure.
            - entry_dict (dict): A dictionary containing information
            - None: Placeholder for future use (currently returns None).

    Structure of `entry_dict`:
        - "csd_code" (str): CSD code (if available) for the crystal structure.
        - "mol_smi" (str): SMILES representation of the molecule.
        - "ccdc_number" (str): CCDC identifier number.
        - "space_group" (str): Space group symbol of the crystal.
        - "spg_num" (int): Space group number.
        - "Z" (int): Number of molecules in the unit cell.
        - "Zprime" (float): Z' value of the crystal.
        - "url" (str): URL link to the CCDC database entry for the crystal.
        - "mol_formula" (str): Molecular formula of the structure.
        - "mol_weight" (float): Molecular weight of the structure.
        - "mol_name" (str): Name of the molecule, typically the CSD code.
        - "l_type" (str): Lattice type of the structure.

    Returns None if the PyXtal structure is invalid (i.e., `xtal.valid` is False).

    Example:
        entry = make_entry_from_pyxtal(xtal_instance)
        ase_atoms, entry_dict, _ = entry

    Notes:
        - The CCDC link is generated using the structure's CCDC number.
    """

    from rdkit import Chem
    from rdkit.Chem.Descriptors import ExactMolWt
    from rdkit.Chem.rdMolDescriptors import CalcMolFormula

    if xtal.valid:
        url0 = "https://www.ccdc.cam.ac.uk/structures/Search?Ccdcid="
        # Create RDKit molecule from SMILES string
        m = Chem.MolFromSmiles(xtal.tag["smiles"])

        # Calculate molecular weight and molecular formula using RDKit
        mol_wt = ExactMolWt(m)
        mol_formula = CalcMolFormula(m)

        # Create a dictionary containing information
        kvp = {
            "csd_code": xtal.tag["csd_code"],
            "mol_smi": xtal.tag["smiles"],
            "ccdc_number": xtal.tag["ccdc_number"],
            "space_group": xtal.group.symbol,
            "spg_num": xtal.group.number,
            "Z": sum(xtal.numMols),
            "Zprime": xtal.get_zprime()[0],
            "url": url0 + str(xtal.tag["ccdc_number"]),
            "mol_formula": mol_formula,
            "mol_weight": mol_wt,
            "mol_name": xtal.tag["csd_code"],
            "l_type": xtal.lattice.ltype,
        }
        # Return the ASE Atoms the entry dictionary, and None as a placeholder
        return (xtal.to_ase(), kvp, None)
    else:
        return None


def make_entry_from_CSD_web(code, number, smiles, name=None):
    """
    make enetry dictionary from csd web https://www.ccdc.cam.ac.uk/structures

    Args:
        code: CSD style letter entry
        number: ccdc number
        smiles: the corresponding molecular smiles
        name: name of the compound
    """

    # xtal = pyxtal(molecular=True)
    #
    # return make_entry_from_pyxtal(xtal)
    raise NotImplementedError("To do in future")


def make_entry_from_CSD(code):
    """
    make entry dictionary from CSD codes

    Args:
        code: a list of CSD codes
    """
    from pyxtal.msg import CSDError

    xtal = pyxtal(molecular=True)
    try:
        xtal.from_CSD(code)
        return make_entry_from_pyxtal(xtal)

    except CSDError as e:
        print("CSDError", code, e.message)
        return None


def make_db_from_CSD(dbname, codes):
    """
    make database from CSD codes

    Args:
        dbname: db file name
        codes: a list of CSD codes
    """
    # open
    db = database(dbname)

    # add structure
    for i, code in enumerate(codes):
        entry = make_entry_from_CSD(code)
        if entry is not None:
            db.add(entry)
            print(i, code)
    return db


class database:
    """
    This is a database class to process crystal data.

    Args:
        db_name: `*.db` format from ase database
    """

    def __init__(self, db_name):
        self.db_name = db_name
        # if not os.path.exists(db_name):
        #    raise ValueError(db_name, 'doesnot exist')

        self.db = connect(db_name, serial=True)
        self.codes = self.get_all_codes()
        self.keys = [
            "csd_code",
            "space_group",
            "spg_num",
            "Z",
            "Zprime",
            "url",
            "mol_name",
            "mol_smi",
            "mol_formula",
            "mol_weight",
            "l_type",
        ]
        self.calculators = [
            "charmm",
            "gulp",
            "ani",
            "dftb_D3",
            "dftb_TS",
        ]

    def vacuum(self):
        self.db.vacuum()

    def get_all_codes(self, group=None):
        """
        Get all codes
        """
        codes = []
        for row in self.db.select():
            if row.csd_code not in codes:
                if group is None:
                    codes.append(row.csd_code)
                else:
                    if row.group == group:
                        codes.append(row.csd_code)
            else:
                print("find duplicate! remove", row.id, row.csd_code)
                self.db.delete([row.id])
        return codes
        # self.codes = codes

    def add(self, entry):
        (atom, kvp, data) = entry
        if kvp["csd_code"] not in self.codes:
            kvp0 = self.process_kvp(kvp)
            self.db.write(atom, key_value_pairs=kvp0, data=data)
            self.codes.append(kvp["csd_code"])

    def add_from_code(self, code):
        entry = make_entry_from_CSD(code)
        if entry is not None:
            self.add(entry)
        else:
            print(f"{code:s} is not a valid entry")

    def process_kvp(self, kvp):
        kvp0 = {}
        for key in self.keys:
            if key in kvp:
                kvp0[key] = kvp[key]
            else:
                print("Error, cannot find ", key, " from the input")
                return None
        return kvp0

    def check_status(self, show=False):
        """
        Check the current status of each entry
        """
        ids = []
        for row in self.db.select():
            if len(row.data.keys()) == len(self.calculators):
                ids.append(row.id)
                if show:
                    row_info = self.get_row_info(id=row.id)
                    self.view(row_info)
            else:
                print(row.csd_code)  # , row.data['charmm_info']['prm'])
        return ids

    def copy(self, db_name, csd_codes):
        """
        copy the entries to another db

        Args:
            db_name: db file name
            csd_codes: list of codes
        """
        if db_name == self.db_name:
            raise RuntimeError("Cannot use the same db file for copy")
        with connect(db_name, serial=True) as db:
            for csd_code in csd_codes:
                row_info = self.get_row_info(code=csd_code)
                (atom, kvp, data) = row_info
                db.write(atom, key_value_pairs=kvp, data=data)

    def view(self, row_info):
        """
        print the summary of benchmark results

        Args:
            row: row object
        """
        from pyxtal.representation import representation

        (atom, kvp, data) = row_info

        # Reference
        xtal = self.get_pyxtal(kvp["csd_code"])
        rep = xtal.get_1D_representation()
        print("\n", kvp["csd_code"], kvp["mol_smi"], xtal.lattice.volume)
        print(rep.to_string() + " reference")

        # calcs
        for key in data:
            calc = key[:-5]
            time = data[key]["time"]

            rep = data[key]["rep"]
            if type(rep[0]) is not list:
                rep = [rep]
            rep = representation(rep, kvp["mol_smi"]).to_string()

            (dv, msd1, msd2) = data[key]["diff"]
            strs = f"{rep:s} {calc:8s} {time / 60:6.2f} {dv:6.3f}"
            if msd1 is not None:
                strs += f"{msd1:6.3f}{msd2:6.3f}"
            print(strs)

    def get_row_info(self, id=None, code=None):
        match = False
        if id is not None:
            for row in self.db.select(id=id):
                match = True
                break
        elif code is not None:
            for row in self.db.select(csd_code=code):
                match = True
                break
        if match:
            kvp = {}
            for key in self.keys:
                kvp[key] = row.key_value_pairs[key]

            data0 = {}
            for calc in self.calculators:
                key = calc + "_info"
                if key in row.data:
                    data0[key] = row.data[key]

            atom = self.db.get_atoms(id=row.id)
            return (atom, kvp, data0)
        else:
            msg = "cannot find the entry from " + id + code
            raise RuntimeError(msg)

    def get_row(self, code):
        for row in self.db.select(csd_code=code):
            return row
        msg = "cannot find the entry from " + code
        raise RuntimeError(msg)

    def get_pyxtal(self, code):
        from pyxtal import pyxtal
        from pyxtal.msg import ReadSeedError
        from pyxtal.util import ase2pymatgen

        row = self.get_row(code)
        atom = self.db.get_atoms(id=row.id)
        # Reference
        pmg = ase2pymatgen(atom)
        smi = row.mol_smi
        smiles = smi.split(".")
        molecules = [smile + ".smi" for smile in smiles]

        xtal = pyxtal(molecular=True)
        try:
            xtal.from_seed(pmg, molecules=molecules)
        except ReadSeedError:
            xtal.from_seed(pmg, molecules=molecules, add_H=True)

        return xtal

    def compute(self, row, work_dir, skf_dir):
        if len(row.data.keys()) < len(self.calculators):
            # not label information, run antechamber
            atom = self.db.get_atoms(id=row.id)
            if "gulp_info" not in row.data:
                # pmg, c_info, g_info = get_parameters(row, atom)
                # row.data = {"charmm_info": c_info, "gulp_info": g_info}
                pass
            else:
                pmg = ase2pymatgen(atom)

            #data = compute(row, pmg, work_dir, skf_dir)
            #self.db.update(row.id, data=data)
            #print("updated the data for", row.csd_code)


class database_topology:
    """
    This is a database class to process atomic crystal data

    Args:
        db_name (str): `*.db` format from ase database
        rank (int): default 0
        size (int): default 1
        ltol (float): lattice tolerance
        stol (float): site tolerance
        atol (float): angle tolerance
        log_file (str): log_file
    """

    def __init__(self, db_name, rank=0, size=1, ltol=0.05, stol=0.05, atol=3,
                 log_file='db.log'):
        self.rank = rank
        self.size = size
        self.db_name = db_name
        self.db = connect(db_name, serial=True)
        self.keys = [
            "space_group_number",
            "pearson_symbol",
            "similarity0",
            "similarity",
            "density",
            "dof",
            "topology",
            "topology_detail",
            "dimension",
            "wps",
            "ff_energy",
            "ff_lib",
            "ff_relaxed",
            "mace_energy",
            "mace_relaxed",
            "dftb_energy",
            "dftb_relaxed",
            "vasp_energy",
            "vasp_relaxed",
        ]
        self.matcher = sm.StructureMatcher(
            ltol=ltol, stol=stol, angle_tol=atol)

        # Define logfile
        self.log_file = log_file
        logging.getLogger().handlers.clear()
        logging.basicConfig(format="%(asctime)s| %(message)s",
                            filename=self.log_file,
                            level=logging.INFO)
        self.logging = logging

    def vacuum(self):
        self.db.vacuum()

    def print_memory_usage(self):
        import psutil
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss / 1024 ** 2
        self.logging.info(f"Rank {self.rank} memory: {mem:.1f} MB")
        print(f"Rank {self.rank} memory: {mem:.1f} MB")

    def get_pyxtal(self, id, use_relaxed=None):
        """
        Get pyxtal based on row_id, if use_relaxed, get pyxtal from ff_relaxed

        Args:
            id (int): row id
            use_relaxed (str): 'ff_relaxed', 'vasp_relaxed'
        """
        from pymatgen.core import Structure
        from pyxtal import pyxtal
        from pyxtal.util import ase2pymatgen

        row = self.db.get(id)  # ; print(id, row.id)
        if use_relaxed is not None:
            if hasattr(row, use_relaxed):
                xtal_str = getattr(row, use_relaxed)
                pmg = Structure.from_str(xtal_str, fmt="cif")
            else:
                print(f"No {use_relaxed} attributes for structure", id)
                atom = self.db.get_atoms(id=id)
                pmg = ase2pymatgen(atom)
        else:
            atom = self.db.get_atoms(id=id)
            pmg = ase2pymatgen(atom)

        xtal = pyxtal()
        try:
            xtal.from_seed(pmg)
            if xtal is not None and xtal.valid:
                for key in self.keys:
                    if hasattr(row, key):
                        setattr(xtal, key, getattr(row, key))
            return xtal
        except:
            print("Cannot load the structure")

    def get_all_xtals(self, include_energy=False):
        """
        Get all pyxtal instances from the current db
        """
        xtals = []
        for row in self.db.select():
            xtal = self.get_pyxtal(id=row.id)
            if xtal is not None:
                if include_energy and hasattr(row, 'vasp_energy'):
                    xtal.energy = row.vasp_energy
                xtals.append(xtal)
        return xtals

    def add_xtal(self, xtal, kvp):
        """
        Add new xtal to the given db
        """
        spg_num = xtal.group.number
        density = xtal.get_density()
        dof = xtal.get_dof()
        wps = [s.wp.get_label() for s in xtal.atom_sites]
        _kvp = {
            "space_group_number": spg_num,
            "pearson_symbol": xtal.get_Pearson_Symbol(),
            "wps": str(wps),
            "density": density,
            "dof": dof,
        }
        kvp.update(_kvp)
        atoms = xtal.to_ase(resort=False)
        self.db.write(atoms, key_value_pairs=kvp)

    def add_strucs_from_db(self, db_file, check=False, tol=0.1, freq=50):
        """
        Add new structures from the given db_file

        Args:
            db_file (str): database path
            tol (float): tolerance in angstrom for symmetry detection
            freq (int): print frequency
        """
        print(f"\nAdding new strucs from {db_file:s}")

        count = 0
        with connect(db_file, serial=True) as db:
            for row in db.select():
                atoms = row.toatoms()
                xtal = pyxtal()
                try:
                    xtal.from_seed(atoms, tol=tol)
                except:
                    xtal = None
                if xtal is not None and xtal.valid:
                    add = self.check_new_structure(xtal) if check else True
                    if add:
                        kvp = {}
                        for key in self.keys:
                            if hasattr(row, key):
                                kvp[key] = getattr(row, key)
                            elif key == "space_group_number":
                                kvp[key] = xtal.group.number
                            elif key == "density":
                                kvp[key] = xtal.get_density()
                            elif key == "dof":
                                kvp[key] = xtal.get_dof()
                            elif key == "wps":
                                kvp[key] == str(s.wp.get_label()
                                                for s in xtal.atom_sites)
                            elif key == "pearson_symbol":
                                kvp[key] = xtal.get_Pearson_Symbol()

                        self.db.write(atoms, key_value_pairs=kvp)
                        count += 1

                if count % freq == 0:
                    print(f"Adding {count:4d} strucs from {db_file:s}")

    def check_new_structure(self, xtal, same_group=True):
        """
        Check if the input xtal already exists in the db

        Args:
            xtal: pyxtal object
            same_group (bool): keep the same group or not
        """

        s_pmg = xtal.to_pymatgen()
        for row in self.db.select():
            if same_group and row.space_group_number != xtal.group.number:
                continue
            ref = self.db.get_atoms(id=row.id)
            ref_pmg = ase2pymatgen(ref)
            if self.matcher.fit(s_pmg, ref_pmg, symmetric=True):
                return False
        return True

    def clean_structures_spg_topology(self, dim=None):
        """
        Clean up the db by removing the duplicate structures
        Here we check the follow criteria
            - same number of atoms
            - same space group
            - same topology
            - same wps

        Args:
            dim (int): wanted dimension
        """

        unique_rows = []
        to_delete = []

        for row in self.db.select():
            unique = True
            # Ignore unwanted dimension
            if dim is not None and hasattr(row, "dimension") and row.dimension != dim:
                # print(row.dimension, dim)
                unique = False
            else:
                for prop in unique_rows:
                    (natoms, spg, wps, topology) = prop
                    if (natoms == row.natoms and spg == row.space_group_number and wps == row.wps) and hasattr(
                        row, "topology"
                    ):
                        if row.topology == "aaa":
                            if row.topology_detail == topology:
                                unique = False
                                break
                        elif row.topology == topology:
                            unique = False
                            break
            if unique:
                if hasattr(row, "topology"):
                    unique_rows.append(
                        (
                            row.natoms,
                            row.space_group_number,
                            row.wps,
                            row.topology if row.topology != "aaa" else row.topology_detail,
                        )
                    )
                else:
                    unique_rows.append(
                        (row.natoms, row.space_group_number, row.wps, None))
            else:
                to_delete.append(row.id)
        if len(to_delete) > 0:
            print(len(to_delete), "structures were deleted", to_delete)
        self.db.delete(to_delete)

    def clean_structures(self, ids=(None, None), dtol=2e-3, etol=1e-3, criteria=None):
        """
        Clean up the db by removing the duplicate structures
        Here we check the follow criteria
        - same number of atoms
        - same density
        - same energy

        Args:
            dtol (float): tolerance of density
            etol (float): tolerance of energy
            criteria (dict): including
        """

        unique_rows = []
        to_delete = []
        ids, xtals = self.select_xtals(ids)
        for id, xtal in zip(ids, xtals):
            row = self.db.get(id)
            xtal = self.get_pyxtal(id)
            unique = True
            if criteria is not None:
                if not xtal.check_validity(criteria, True):
                    unique = False
                    print(
                        "Found unsatisfied criteria",
                        row.id,
                        row.space_group_number,
                        row.wps,
                    )

                if unique and (
                    "MAX_energy" in criteria and hasattr(
                        row, "ff_energy") and row.ff_energy > criteria["MAX_energy"]
                ):
                    unique = False
                    print(
                        "Unsatisfied energy",
                        row.id,
                        row.ff_energy,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "MAX_similarity" in criteria
                    and hasattr(row, "similarity")
                    and row.similarity > criteria["MAX_similarity"]
                ):
                    unique = False
                    print(
                        "Unsatisfied similarity",
                        row.id,
                        row.similarity,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "BAD_topology" in criteria
                    and hasattr(row, "topology")
                    and row.topology[:3] in criteria["BAD_topology"]
                ):
                    unique = False
                    print(
                        "Unsatisfied topology",
                        row.id,
                        row.topology,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "BAD_dimension" in criteria
                    and hasattr(row, "dimension")
                    and row.dimension in criteria["BAD_dimension"]
                ):
                    unique = False
                    print(
                        "Unsatisfied dimension",
                        row.id,
                        row.topology,
                        row.space_group_number,
                        row.wps,
                    )

            if unique:
                for prop in unique_rows:
                    (natoms, spg, wps, den, ff_energy) = prop
                    if natoms == row.natoms and spg == row.space_group_number and wps == row.wps:
                        if hasattr(row, "ff_energy") and ff_energy is not None:
                            if abs(row.ff_energy - ff_energy) < etol:
                                unique = False
                                break
                        elif abs(den - row.density) < dtol:
                            unique = False
                            break

            if unique:
                if hasattr(row, "ff_energy"):
                    unique_rows.append(
                        (
                            row.natoms,
                            row.space_group_number,
                            row.wps,
                            row.density,
                            row.ff_energy,
                        )
                    )
                else:
                    unique_rows.append(
                        (row.natoms, row.space_group_number, row.wps, row.density, None))
            else:
                to_delete.append(row.id)
        print(len(to_delete), "structures were deleted", to_delete)
        self.db.delete(to_delete)

    def clean_structures_pmg(self, ids=(None, None), min_id=None, dtol=5e-2, criteria=None):
        """
        Clean up the db by removing the duplicate structures.
        Here we check the follow criteria same density and pymatgen matcher.
        The criteria should look like the following,

        {'CN': {'C': 3},
         'cutoff': 1.8,
         'MAX_energy': -8.00,
         #'MAX_similarity': 0.2,
         'BAD_topology': ['hcb'],
         'BAD_dimension': [0, 2],
        }

        Args:
            dtol (float): tolerance of density
            criteria (dict): including
        """

        unique_rows = []
        to_delete = []

        ids, xtals = self.select_xtals(ids)
        if min_id is None:
            min_id = min(ids)

        for id, xtal in zip(ids, xtals):
            row = self.db.get(id)
            xtal = self.get_pyxtal(id)
            unique = True

            if id > min_id and criteria is not None:
                if not xtal.check_validity(criteria, True):
                    unique = False
                    print(
                        "Found unsatisfied criteria",
                        row.id,
                        row.space_group_number,
                        row.wps,
                    )

                if unique and (
                    "MAX_energy" in criteria and hasattr(
                        row, "ff_energy") and row.ff_energy > criteria["MAX_energy"]
                ):
                    unique = False
                    print(
                        "Unsatisfied energy",
                        row.id,
                        row.ff_energy,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "MAX_similarity" in criteria
                    and hasattr(row, "similarity")
                    and row.similarity > criteria["MAX_similarity"]
                ):
                    unique = False
                    print(
                        "Unsatisfied similarity",
                        row.id,
                        row.similarity,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "BAD_topology" in criteria
                    and hasattr(row, "topology")
                    and row.topology[:3] in criteria["BAD_topology"]
                ):
                    unique = False
                    print(
                        "Unsatisfied topology",
                        row.id,
                        row.topology,
                        row.space_group_number,
                        row.wps,
                    )
                if unique and (
                    "BAD_dimension" in criteria
                    and hasattr(row, "dimension")
                    and row.dimension in criteria["BAD_dimension"]
                ):
                    unique = False
                    print(
                        "Unsatisfied dimension",
                        row.id,
                        row.topology,
                        row.space_group_number,
                        row.wps,
                    )

            if unique and id > min_id:
                for prop in unique_rows:
                    (rowid, den) = prop
                    if abs(den - row.density) < dtol:
                        ref_pmg = xtal.to_pymatgen()
                        s_pmg = ase2pymatgen(self.db.get_atoms(id=rowid))
                        # , symmetric=True):
                        if self.matcher.fit(s_pmg, ref_pmg):
                            print(
                                "Found duplicate",
                                row.id,
                                row.space_group_number,
                                row.wps,
                            )
                            unique = False
                            break
            if unique:
                unique_rows.append((row.id, row.density))
            else:
                to_delete.append(row.id)
        print(len(to_delete), "structures were deleted", to_delete)
        self.db.delete(to_delete)

    def get_max_id(self):
        """
        Get the maximum row id
        """
        max_id = None
        for row in self.db.select():
            if max_id is None or row.id > max_id:
                max_id = row.id + 1
        return max_id

    def select_xtals(self, ids, N_atoms=(None, None), overwrite=False, attribute=None, use_relaxed=None):
        """
        Extract xtals based on attribute name.
        Mostly called by update_row_energy

        Args:
            ids:
            N_atoms:
            overwrite:
            atttribute:
            use_relaxed (str): 'ff_relaxed' or 'vasp_relaxed'
        """
        (min_id, max_id) = ids
        if min_id is None: min_id = 1
        if max_id is None: max_id = self.get_max_id()

        (min_atoms, max_atoms) = N_atoms
        if min_atoms is None: min_atoms = 1
        if max_atoms is None: max_atoms = 5000

        ids, xtals = [], []
        for row in self.db.select():
            if overwrite or attribute is None or not hasattr(row, attribute):
                if min_id <= row.id <= max_id and min_atoms < natoms <= max_atoms:
                    xtal = self.get_pyxtal(row.id, use_relaxed)
                    ids.append(row.id)
                    xtals.append(xtal)
                    if len(xtals) % 100 == 0:
                        print("Loading xtals from db", len(xtals))
        return ids, xtals

    def select_xtal(self, ids, N_atoms=(None, None), overwrite=False, attribute=None, use_relaxed=None):
        """
        Lazy extraction of select xtals

        Args:
            ids:
            N_atoms:
            overwrite:
            atttribute:
            use_relaxed (str): 'ff_relaxed' or 'vasp_relaxed'
        """
        (min_id, max_id) = ids
        if min_id is None: min_id = 1
        if max_id is None: max_id = self.get_max_id()

        (min_atoms, max_atoms) = N_atoms
        if min_atoms is None: min_atoms = 1
        if max_atoms is None: max_atoms = 5000

        ids, xtals = [], []
        for row in self.db.select():
            if overwrite or attribute is None or not hasattr(row, attribute):
                id, natoms = row.id, row.natoms
                if min_id <= id <= max_id and \
                    min_atoms < natoms <= max_atoms \
                    and id % self.size== self.rank:

                    xtal = self.get_pyxtal(id, use_relaxed)
                    yield id, xtal

    def update_row_energy(
        self,
        calculator='GULP',
        ids=(None, None),
        N_atoms=(None, None),
        ncpu=1,
        criteria=None,
        symmetrize=False,
        overwrite=False,
        write_freq=100,
        ff_lib='reaxff',
        steps=250,
        use_relaxed=None,
        cmd=None,
        calc_folder=None,
        skf_dir=None,
    ):
        """
        Update the row energy in the database for a given calculator.

        Args:
            calculator (str): 'GULP', 'MACE', 'VASP', 'DFTB'
            ids (tuple): A tuple specifying row IDs to update (e.g., (0, 100)).
            ncpu (int): number of parallel processes
            criteria (dict, optional): Criteria when selecting structures.
            symmetrize (bool): symmetrize the structure before calculation
            overwrite (bool): overwrite the existing energy attributes.
            write_freq (int): frequency to update db for ncpu=1
            ff_lib (str): Force field to use for GULP ('reaxff' by default).
            steps (int): Number of optimization steps for DFTB (default is 250).
            use_relaxed (str, optional): Use relaxed structures (e.g. 'ff_relaxed')
            cmd (str, optional): Command for VASP calculations
            calc_folder (str, optional): calc_folder for GULP/VASP calculations
            skf_dir (str, optional): Directory for DFTB potential files

        Functionality:
            Using the selected calculator, it updates the energy rows of the
            database. If `ncpu > 1`, run in parallel; otherwise in serial.

        Calculator Options:
            - 'GULP': Uses a force field (e.g., 'reaxff').
            - 'MACE': Uses the MACE calculator.
            - 'DFTB': Uses DFTB+ with symmetrization options.
            - 'VASP': Uses VASP, with a specified command (`cmd`).
        """

        label = calculator.lower() + "_energy"
        if calc_folder is None:
            calc_folder = calculator.lower() + "_calc"

        if calculator != 'MACE':
            #self.logging.info("make new folders", calc_folder, os.getpwd())
            os.makedirs(calc_folder, exist_ok=True)

        # Generate structures for calculation
        generator = self.select_xtal(ids, N_atoms, overwrite, label, use_relaxed)

        # Set up arguments for the chosen calculator
        args_up = []
        if calculator == 'GULP':
            args = [calculator, ff_lib, calc_folder, criteria]
            args_up = [ff_lib]
        elif calculator == 'MACE':
            args = [calculator, steps, criteria]
        elif calculator == 'DFTB':
            args = [calculator, skf_dir, steps, symmetrize, criteria]
        elif calculator == 'VASP':
            args = [calculator, calc_folder, cmd, criteria]
        else:
            raise ValueError(f"Unsupported calculator: {calculator}")

        # Perform calculation serially or in parallel
        self.logging.info(f"Rank-{self.rank} row_energy {calculator} {self.db_name}")
        if ncpu == 1:
            self.update_row_energy_serial(generator, write_freq, args, args_up)
        else:
            self.update_row_energy_mproc(ncpu, generator, args, args_up)
        self.logging.info(f"Rank-{self.rank} complete update_row_energy")

    def update_row_energy_serial(self, generator, write_freq, args, args_up):
        """
        Perform a serial update of row energies

        Args:
            generator (generator): Yielding tuples of (id, xtal), where:
                - `id` (int): Unique identifier for the structure.
                - `xtal` (object): pyxtal instance.
            write_freq (int): Frequency to update the database.
            args (list): Additional arguments to the function `opt_single`.
            args_up (list): Additional arguments for function `_update_db`.

        Functionality:
            It iterates over structures provided by `generator`,
            optimizes them using `opt_single`, and collects results that have
            converged (`status == True`). Once the number of results
            reaches `write_freq`, it updates the database.
        """
        results = []
        for id, xtal in generator:
            self.logging.info(f"Processing {id} {xtal.lattice} {args[0]}")
            print(f"Processing {id} {xtal.lattice} {args[0]}")
            res = opt_single(id, xtal, *args)
            (xtal, eng, status) = res
            if status:
                results.append((id, xtal, eng))
            if len(results) >= write_freq:
                self._update_db(results, args[0], *args_up)
                results = []
                self.print_memory_usage()
        if len(results) > 0:
            self._update_db(results, args[0], *args_up)

    def update_row_energy_mproc(self, ncpu, generator, args, args_up):
        """
        Perform parallel row energy updates by optimizing atomic structures.

        Args:
            ncpu (int): Number of CPUs to use for parallel processing.
            generator (generator): yielding tuples of (id, xtal), where:
                - `id` (int): Unique identifier for the structure.
                - `xtal` (object): pyxtal instance.
            args (list): Additional arguments passed to `call_opt_single`.
                - Typically includes a calculator or potential parameters.
            args_up (list): Additional arguments for function `_update_db`.

        Functionality:
            This function distributes the structures across multiple CPUs
            using `multiprocessing.Pool`. It creates chunks (based on `ncpu`),
            and process them in parallel by calling `call_opt_single`.
            Successful results are periodically written to the database.
            The function also prints memory usage after each database update.

        Parallelization Process:
            - The `Pool` is initialized with `ncpu` processes.
            - Structures are divided into chunks with the `chunkify` function.
            - Each chunk is processed by `call_opt_single` via the pool.
            - Successful results are periodically written to the database.
            - The pool is closed and joined after processing is complete.
        """
        from multiprocessing import Pool

        self.logging.info(f"Parallel optimizations {ncpu}")
        pool = Pool(processes=ncpu,
                    initializer=setup_worker_logger,
                    initargs=(self.log_file,))

        def chunkify(generator, chunk_size):
            chunk = []
            for item in generator:
                chunk.append(item)
                if len(chunk) == chunk_size:
                    yield chunk
                    chunk = []
            if chunk:
                yield chunk

        for chunk in chunkify(generator, ncpu*10):
            myargs = []
            for _id, xtal in chunk:
                if xtal is not None:
                    myargs.append(tuple([_id, xtal] + args))

            results = []
            self.logging.info(f"Start minicycle: {myargs[0][0]}-{myargs[-1][0]}")
            for result in pool.imap_unordered(call_opt_single,
                                              myargs,
                                              chunksize=1):
                if result is not None:
                    (myid, xtal, eng) = result
                    if eng is not None:
                        results.append(result)
                        numIons = sum(xtal.numIons)
                        count = len(results)
                        self.logging.info(f"Add {myid:4d} {eng:.3f} *{numIons} {count}")

                # Only do frequent update for slow calculator VASP
                if len(results) >= ncpu and args[0] == 'VASP':
                    self._update_db(results, args[0], *args_up)
                    self.logging.info(f"Finish minibatch: {len(results)}")
                    self.print_memory_usage()
                    results = []

            self.logging.info(f"Done  minicycle: {myargs[0][0]}-{myargs[-1][0]}")

            # After the loop, handle the remaining results
            if results:
                self.logging.info(f"Start  Update db: {len(results)}")
                self._update_db(results, args[0], *args_up)
                self.logging.info(f"Finish Update db: {len(results)}")

        pool.close()
        pool.join()

    def _update_db(self, results, calc, *args):
        """
        Update db with the calculation_results
        https://wiki.fysik.dtu.dk/ase/ase/db/db.html#writing-and-updating-many-rows-efficiently

        Args:
            results: list of (id, xtal, eng) tuples
            calc (str): calculator
        """
        #self.logging.info(f"====================Update db: {len(results)}")
        if calc == 'GULP':
            ff_lib = args[0]

        with self.db:
            for result in results:
                (id, xtal, eng) = result
                if xtal is not None:
                    if calc == 'GULP':
                        self.db.update(id,
                               ff_energy=eng,
                               ff_lib=ff_lib,
                               ff_relaxed=xtal.to_file())
                    elif calc == 'MACE':
                        self.db.update(id,
                               mace_energy=eng,
                               mace_relaxed=xtal.to_file())
                    elif calc == 'VASP':
                        self.db.update(id,
                                vasp_energy=eng,
                                vasp_relaxed=xtal.to_file())
                    elif calc == 'DFTB':
                        self.db.update(id,
                                dftb_energy=eng,
                                dftb_relaxed=xtal.to_file())
                    #self.logging.info(f'update_db_{calc}, {id}')

    def update_row_topology(self, StructureType="Auto", overwrite=True, prefix=None, ref_dim=3):
        """
        Update row topology base on the CrystalNets.jl

        Args:
            StructureType (str): 'Zeolite', 'MOF' or 'Auto'
            overwrite (bool): remove the existing attributes
            prefix (str): prefix for tmp cif file
            ref_dim (int): desired dimension
        """
        try:
            import juliacall
        except:
            raise RuntimeError(
                "Cannot load JuliaCall, Plz enable it before running")

        def parse_topology(topology_info):
            """
            Obtain the dimensionality and topology name
            """
            dim = 0
            name = ""
            detail = "None"
            for i, x in enumerate(topology_info):
                (d, n) = x
                if d > dim:
                    dim = d
                tmp = n.split(",")[0]
                if tmp.startswith("UNKNOWN"):
                    # tuple(int(num) for num in tmp[7:].split())
                    detail = tmp[7:]
                    tmp = "aaa"
                elif tmp.startswith("unstable"):
                    tmp = "unstable"
                name += tmp
                if i + 1 < len(topology_info):
                    name += "-"
            return dim, name, detail

        jl = juliacall.newmodule("MOF_Builder")
        jl.seval("using CrystalNets")
        jl.CrystalNets.toggle_warning(False)  # to disable warnings
        jl.CrystalNets.toggle_export(False)  # to disable exports
        if StructureType == "Zeolite":
            option = jl.CrystalNets.Options(structure=jl.StructureType.Zeolite)
        elif StructureType == "MOF":
            option = jl.CrystalNets.Options(structure=jl.StructureType.MOF)
        else:
            option = jl.CrystalNets.Options(structure=jl.StructureType.Auto)

        cif_file = prefix + ".cif" if prefix is not None else "tmp.cif"

        for row in self.db.select():
            if overwrite or not hasattr(row, "topology"):
                atoms = self.db.get_atoms(row.id)
                atoms.write(cif_file, format="cif", parallel=False)

                # Call crystalnet.jl
                result = jl.determine_topology(cif_file, option)
                # print(result)
                results = list(result) if len(result) > 1 else [result[0]]
                try:
                    topo = []
                    for res in results:
                        # topology for SingleNodes
                        name = str(res[0])
                        if res[1] > 1:
                            name += "(" + str(res[1]) + ")"
                        genome = res[0][jl.Clustering.Auto]
                        dim = jl.ndims(jl.CrystalNets.PeriodicGraph(genome))
                        topo.append((dim, name))

                    # The maximum dimensionality and topology name
                    dim, name, detail = parse_topology(topo)
                except:
                    dim, name, detail = 3, "error", "error"
                if dim == ref_dim:
                    print(
                        "Updating Topology",
                        row.space_group_number,
                        row.wps,
                        dim,
                        name,
                        detail[:10],
                    )
                # Unknown will be labeled as aaa
                self.db.update(row.id, topology=name,
                               dimension=dim, topology_detail=detail)
            # else:
            #    print("Existing Topology", row.topology)

    def update_db_description(self):
        """
        update db description based on robocrys
        Call robocrys: https://github.com/hackingmaterials/robocrystallographer
        """
        from robocrys import StructureCondenser, StructureDescriber

        condenser = StructureCondenser()
        describer = StructureDescriber()

        for row in self.db.select():
            if not hasattr(row, "description"):
                atoms = self.db.get_atoms(row.id)
                pmg = ase2pymatgen(atoms)
                try:
                    condensed_structure = condenser.condense_structure(pmg)
                    description = describer.describe(condensed_structure)
                except:
                    description = "N/A"

                self.db.update(row.id, description=description)
                print("\n======Updating\n", description)
            else:
                print("\n======Existing\n", row.description)

    def export_structures(
        self,
        fmt="vasp",
        folder="mof_out",
        criteria=None,
        sort_by="similarity",
        overwrite=True,
        cutoff=None,
        use_relaxed=None,
    ):
        """
        export structures from database according to the given criterion

        Args:
            fmt (str): 'vasp' or 'cif'
            folder (str): 'path of output folders'
            criteria (dict): check the validity with dict
            sort_by (str): sort by which attribute
            overwrite (bool): remove the existing folder
            cutoff (int): the maximum number of structures for export
        """

        import shutil

        if cutoff is None:
            cutoff = self.db.count()
        if not os.path.exists(folder):
            os.makedirs(folder)
        else:
            if overwrite:
                shutil.rmtree(folder)
                os.makedirs(folder)

        keys = [
            "id",
            "pearson_symbol",
            "space_group_number",
            "density",
            "dof",
            "similarity",
            "ff_energy",
            "vasp_energy",
            "mace_energy",
            "topology",
        ]
        properties = []
        for row in self.db.select():
            spg = row.space_group_number
            den = row.density
            dof = row.dof
            ps = row.pearson_symbol
            sim = float(row.similarity) if hasattr(
                row, "similarity") and row.similarity is not None else None
            top = row.topology if hasattr(row, "topology") else None
            ff_eng = float(row.ff_energy) if hasattr(
                row, "ff_energy") else None
            vasp_eng = float(row.vasp_energy) if hasattr(
                row, "vasp_energy") else None
            mace_eng = float(row.mace_energy) if hasattr(
                row, "mace_energy") else None
            properties.append([row.id, ps, spg, den, dof,
                              sim, ff_eng, vasp_eng, mace_eng,
                              top])

        dicts = {}
        for i, key in enumerate(keys):
            if properties[0][i] is not None:
                dicts[key] = [prop[i] for prop in properties]

        if sort_by in keys:
            col = keys.index(sort_by)  # + 1
        else:
            print("supported attributes", keys)
            raise ValueError("Cannot sort by", sort_by)

        print(f"====Exporting {len(properties)} structures")
        properties = [prop for prop in properties if prop[col] is not None]
        sorted_properties = sorted(properties, key=lambda x: x[col])

        for entry in sorted_properties[:cutoff]:
            [id, ps, spg, den, dof, sim, ff_eng, vasp_eng, mace_eng, top] = entry
            id = int(id)
            spg = int(spg)
            sim = float(sim)
            den = float(den)
            dof = int(dof)
            if vasp_eng is not None:
                eng = float(vasp_eng)
            elif mace_eng is not None:
                eng = float(mace_eng)
            elif ff_eng is not None:
                eng = float(ff_eng)
            else:
                eng = None
            if True:
            #try:
                xtal = self.get_pyxtal(id, use_relaxed)
                number, symbol = xtal.group.number, xtal.group.symbol.replace(
                    "/", "")
                # convert to the desired subgroup representation if needed
                #if number != spg:
                #    paths = xtal.group.path_to_subgroup(spg)
                #    xtal = xtal.to_subgroup(paths)
                #    number, symbol = (
                #        xtal.group.number,
                #        xtal.group.symbol.replace("/", ""),
                #    )

                label = os.path.join(
                    folder,
                    f"{id:d}-{xtal.get_Pearson_Symbol():s}-{number:d}-{symbol:s}",
                )

                status = xtal.check_validity(
                    criteria, True) if criteria is not None else True
            #except:
            #    status = False
            #    label = "Error"

            if status:
                try:
                    # if True:
                    xtal.set_site_coordination()
                    for s in xtal.atom_sites:
                        _l, _sp, _cn = s.wp.get_label(), s.specie, s.coordination
                        label += f"-{_l:s}-{_sp:s}{_cn:d}"
                    label += f"-S{sim:.3f}"
                    if len(label) > 40:
                        label = label[:40]
                except:
                    print("Problem in setting site coordination")

                if den is not None:
                    label += f"-D{abs(den):.2f}"
                if eng is not None:
                    label += f"-E{abs(eng):.3f}"
                if top is not None:
                    label += f"-T{top:s}"
                # if sim is not None: label += '-S{:.2f}'.format(sim)

                print("====Exporting:", label)
                if fmt == "vasp":
                    xtal.to_file(label + ".vasp", fmt="poscar")
                elif fmt == "cif":
                    xtal.to_file(label + ".cif")
            else:
                print("====Skippng:  ", label)

        return dicts

    def get_label(self, i):
        if i < 10:
            folder = f"cpu00{i}"
        elif i < 100:
            folder = f"cpu0{i}"
        else:
            folder = f"cpu0{i}"
        return folder

    def get_db_unique(self, db_name=None, prec=3):
        """
        Get a db file with only unique structures
        with the following identical attributes:
        (topology, ff_energy)

        Args:
            db_name (str): filename for the new db
            prec (int): ff_energy precision for the round number
        """

        print(f"The {self.db_name:s} has {self.db.count():d} strucs")
        if db_name is None:
            db_name = self.db_name[:-3] + "_unique.db"
        if os.path.exists(db_name):
            os.remove(db_name)

        unique_props = {}  # Using a dictionary to store unique properties
        for row in self.db.select():
            if hasattr(row, "topology") and hasattr(row, "ff_energy"):
                top, top_detail = row.topology, row.topology_detail
                dof, ff_energy = row.dof, round(row.ff_energy, prec)
                prop_key = (top, top_detail, ff_energy)
                # A dictionary lookup
                if prop_key in unique_props:
                    _id, _dof = unique_props[prop_key]
                    if dof < _dof:
                        print("Updating", row.id, top, ff_energy)
                        unique_props[prop_key] = (row.id, dof)
                    else:
                        print("Duplicate", row.id, top, ff_energy)
                else:
                    print("Adding", row.id, top, ff_energy)
                    unique_props[prop_key] = (row.id, dof)

        ids = [unique_props[key][0] for key in unique_props.keys()]
        with connect(db_name, serial=True) as db:
            for id in ids:
                row = self.db.get(id)
                kvp = {}
                for key in self.keys:
                    if hasattr(row, key):
                        kvp[key] = getattr(row, key)
                db.write(row.toatoms(), key_value_pairs=kvp)
        print(f"Created {db_name:s} with {db.count():d} strucs")
        return db.count()

    def check_overlap(self, reference_db, etol=2e-3, verbose=True):
        """
        Check the overlap w.r.t the reference database

        Args:
            reference_db (str): path of reference database
            etol (float): energy tolerence to distinguish the identical structure
            verbose (bool): whether or not print out details
        """

        db_ref = database_topology(reference_db, log_file = self.log_file)
        print(f"\nCurrent   database {self.db_name}: {self.db.count()}")
        print(f"Reference database {db_ref.db_name}: {db_ref.db.count()}")

        ref_data = []
        for row in db_ref.db.select():
            if hasattr(row, "topology") and hasattr(row, "ff_energy"):
                ref_data.append(
                    (row.topology, row.topology_detail, row.ff_energy))

        overlaps = []
        for row in self.db.select():
            if hasattr(row, "topology") and hasattr(row, "ff_energy"):
                for ref in ref_data:
                    (top, top_detail, ff_energy) = ref
                    if (
                        row.topology == top
                        and row.topology_detail == top_detail
                        and abs(row.ff_energy - ff_energy) < etol
                    ):
                        # strs = 'Find {:4d} {:6s}'.format(row.id, row.pearson_symbol)
                        # strs += ' {:12s} {:10.3f}'.format(row.topology, row.ff_energy)
                        # print(strs)
                        overlaps.append(
                            (
                                row.id,
                                row.pearson_symbol,
                                row.dof,
                                row.topology,
                                row.ff_energy,
                            )
                        )
                        break
        strs = f"\nThe number of overlap is: {len(overlaps):d}"
        strs += f"/{self.db.count():d}/{db_ref.db.count():d}"
        print(strs)
        sorted_overlaps = sorted(overlaps, key=lambda x: x[-1])
        if verbose:
            for entry in sorted_overlaps:
                print("{:4d} {:6s} {:4d} {:20s} {:10.3f}".format(*entry))

        return overlaps

    def print_info(self, excluded_ids=None, cutoff=100):
        """
        Print out the summary of the database based on the calculated energy
        Mostly used to quickly view the most interesting low-energy structures.
        Todo: show vasp_energy if available

        Args:
            excluded_ids (list): list of unwanted row ids
            cutoff (int): the cutoff value for the print
        """
        if excluded_ids is None:
            excluded_ids = []
        print(f"\nCurrent database {self.db_name}: {self.db.count()}")
        output = []
        for row in self.db.select():
            if row.id not in excluded_ids and hasattr(row, "topology") and hasattr(row, "ff_energy"):
                output.append(
                    (
                        row.id,
                        row.pearson_symbol,
                        row.dof,
                        row.topology,
                        row.ff_energy,
                    )
                )

        sorted_output = sorted(output, key=lambda x: x[-1])
        for entry in sorted_output[:cutoff]:
            print("{:4d} {:6s} {:4d} {:20s} {:10.3f}".format(*entry))

        strs = f"Showed structures: {len(sorted_output)}/{self.db.count()}"
        print(strs)

    def plot_histogram(self, prop, ax=None, filename=None, xlim=None, nbins=20):
        """
        Plot the histogram of a specified row property.

        Args:
            prop (str): The name of the property to plot (e.g., 'ff_energy').
            ax (matplotlib.axes.Axes, optional): Pre-existing axis to plot on.
                                                 If None, a new ax will be created.
            filename (str, optional): Path to save the plot (e.g., 'plot.png').
                                      If None, the plot will not be saved.
            xlim (tuple, optional): Limits for the x-axis (e.g., (0, 10)).
                                    If None, the x-axis will scale automatically.
            nbins (int, optional): Number of bins for the histogram. Default is 20.

        Returns:
            matplotlib.axes.Axes: The axis object with the histogram plotted.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            f, ax = plt.subplots()

        # Get the properties from the database
        props = self.get_properties(prop)

        # Check if there are values to plot
        if not props:
            raise ValueError(f"No rows contain the property '{prop}'.")

        ax.hist(props, nbins, density=True, alpha=0.75)

        # Set x-axis limits if provided
        if xlim is not None:
            ax.set_xlim(xlim)

        ax.set_xlabel(prop)

        # Save the plot if a filename is provided
        if filename is not None:
            plt.savefig(filename)

        return ax

    def get_properties(self, prop):
        """
        Retrieve a list of specific property values from the database rows.

        Args:
            prop (str): The property name to retrieve (e.g., 'ff_energy')

        Returns:
            list: A list of property values for rows that have the specified property.
                  If a row does not contain the property, it is ignored.

        Raises:
            Warning: If no rows in the database contain the specified property.
        """

        props = []

        # Loop through all rows in the database and collect the property values
        for row in self.db.select():
            if hasattr(row, prop):
                props.append(getattr(row, prop))

        # Print summary of rows
        name, count = self.db_name, self.db.count()
        print(f"Database {name} has {prop}: {len(props)}/{count}")

        # Warn if no properties were found
        if count == 0:
            raise Warning(
                f"No rows in the database contain the property '{prop}'.")

        return props


if __name__ == "__main__":
    # open
    if False:
        db = database("test.db")
        print("Total number of entries", len(db.codes))

        # view structure
        c = db.get_pyxtal("HXMTAM")
        print(c)
    if False:
        db = database_topology("../MOF-Builder/C-sp2/sp2-sacada-0506.db")
        # xtal = db.get_pyxtal(1)
        # print(xtal)
        # db.add_xtal(xtal, kvp={'similarity': 0.1})

        # db.update_row_ff_energy(ids=(0, 2), overwrite=True)
        # db.update_row_ff_energy(ncpu=2, ids=(2, 20), overwrite=True)
        # brew install coreutils to get timeout in maca
        # os.environ['ASE_DFTB_COMMAND'] = 'timeout 1m /Users/qzhu8/opt/dftb+/bin/dftb+ > PREFIX.out'
        os.environ["ASE_DFTB_COMMAND"] = "/Users/qzhu8/opt/dftb+/bin/dftb+ > PREFIX.out"
        skf_dir = "/Users/qzhu8/GitHub/MOF-Builder/3ob-3-1/"
        # db.update_row_dftb_energy(skf_dir, ncpu=1, ids=(0, 2), overwrite=True)
        db.update_row_dftb_energy(
            skf_dir, ncpu=1, ids=(17, 17), overwrite=True)

    db = database_topology("total.db")
    db.get_db_unique()
    db1 = database_topology("sp2_sacada.db")
    db1.get_db_unique()
    db = database_topology("total_unique.db")
    db.check_overlap("sp2_sacada_unique.db")
    db1.export_structures(folder="mof_out_sacada")
